#!/usr/bin/perl
###
### http://bach.cs.kobe-u.ac.jp/metcha/
### Metcha Search Engine 1.7
### Copyright: Naoyuki Tamura (tamura@kobe-u.ac.jp)
### 1999/06/25 version 1.0 released
### 1999/07/03 version 1.1 released
### 2000/03/17 version 1.2 released
### 2000/10/25 version 1.3 released
### 2001/01/31 version 1.4 released
### 2001/05/07 version 1.5 released
### 2001/09/18 version 1.6 released
### 2001/12/20 version 1.7 released
### This program is distributed as a free software for non-profit use.
###
use POSIX;
use Socket;
use Getopt::Std;
require 'jcode.pl';

# $log = '/home/httpd/cgi-bin/metcha.log';
$log = '/home/httpd/cgi-bin/log/metcha%.log';
# $logmod = 0666;
$toppage = 'http://bach.cs.kobe-u.ac.jp/metcha/';
$metcha =  '/cgi-bin/metcha.cgi';
$helppage = '/metcha/help.html';
$bmpage = '/metcha/bookmarks.html';
$bmscript = '/metcha/bookmarks.js';
$maintainer = '<a href="/tamura-jp.html" target=_top>Naoyuki Tamura</a>';
$from = 'tamura@kobe-u.ac.jp';
# $re_allow_ip = '^133\.30\.';
# $proxy_server = 'proxy.xxx.yy.jp';
# $proxy_port = 8080;
# $no_proxy = 'xxx\.yy\.jp$';
$max_age = 600;

$agent_name = "Metcha/1.7 ($toppage)";
$programmer =
    '<a href="http://bach.cs.kobe-u.ac.jp/tamura-jp.html" target=_top>'
    . 'Naoyuki Tamura</a>';
$| = 1;				# No buffering

&main();
exit 0;

sub main {
    my ($eng, $n, $t0, $t, $file);
    my (@hrefs);
    &init_params;	        # Initialize parameters
    if ($query =~ /^\s*$/) {
	print "Location: $toppage\n\n";
	return 1;
    }
    &print_header;		# Print header
    if ($cgi) {
	&write_log_1($logfile, $id, "[$current_date]");
    }
    if ($cgi && $re_allow_ip && $ENV{'REMOTE_ADDR'} !~ /$re_allow_ip/) {
	print "You are not authorized to use this service.\n";
	print "</body></html>\n";
	return -1;
    }

    ## Generate requests
    print "Searching\n";
    foreach $eng (@engines) {
	&init_engine($eng, $max_hits);
	print "[<a href=\"http://$server{$eng}$request{$eng}\"$target>";
	print "$eng</a>]\n";
    }
    print "<br>\n";

    ## Send requests
    $n = @engines;
    print "<font size=-1 color=#808080>";
    print "Connecting to $n engines (seconds):\n";
    foreach $eng (@engines) {
	$t0 = time;
	if ($debug ne 'l') {
	    &request($eng);
	}
	$t = time - $t0;
	if ($status{$eng} ne 'err') {
	    print "$eng=$t\n";
	} else {
	    print "$eng=x\n";
	}
    }
    print "</font><br>\n";

    $engines = grep(($status{$_} ne 'err'), @engines);

    ## Receive from all engines
    print "<font size=-1 color=#808080>";
    print "Receiving from $engines engines (seconds):\n";
    if ($debug ne 'l') {
	&receive_all(time);
    } else {
	foreach $eng (@engines) {
	    $file = "$eng.log";
	    &load_content($eng, $file);
	    $status{$eng} = 'ok';
	    print "$eng=0\n";
	}
    }
    print "</font><br>\n";
    if ($debug eq 's') {
	foreach $eng (@engines) {
	    $file = "$eng.log";
	    unlink $file;
	    next if $status{$eng} eq 'err';
	    &save_content($eng, $file);
	}
    }

    ## Extract URLs
    print "<font size=-1 color=#808080>";
    print "Extracting from $engines engines (URLs):\n";
    foreach $eng (@engines) {
	next if $status{$eng} eq 'err';
	&extract($eng);
	print "$eng=$count{$eng}\n";
    }
    print "</font><br>\n";

    undef %content;

    ## Calculate total score
    &calc_total;
    $eng = 'TOTAL';
    $status{$eng} = 'ok';
    $time{$eng} = time - $start_time;
    print ": $count{$eng} ڡ ($time{$eng} )<br>\n";
    print "<hr>\n";

    ## Order by the total score
    @hrefs = sort by_total keys %total;
    if ($max_show > 0 && @hrefs > $max_show) {
	splice(@hrefs, $max_show);
    }
    $c = 0;
    foreach $href (@hrefs) {
	$total_rank{$href} = ++$c;
    }

    ## List Pages
    print "<form name=\"results\">\n";

    ## List Good Pages
    $good = 0;
    if (%good) {
	print "<!-- begin good pages -->\n";
	&list_good_pages;
	print "<!-- end good pages -->\n";
	print "<hr>\n";
    }

    ## List Pages
    print "<!-- begin results -->\n";
    if (! $list_by_site) {
	# List by page
	&list_by_page(@hrefs);
    } else {
	# List by site
	&list_by_site(@hrefs);
    }
    print "<!-- end results -->\n";
    print "</form>\n";

    ## Print links
    &print_links;

    &print_trailer;			# Print trailer
    if ($cgi) {
	$t = time - $start_time;
	&write_log_2($logfile, $id, "[$t]");
    }
    return 0;
}

##
## Initialize Parameters
##
sub init_params {
    $start_time = time;
    $id = "$start_time:$$";
    $current_date = scalar(localtime($start_time));
    $logfile = $log;
    if ($logfile =~ /%/) {
	@_ = localtime($start_time);
	my $ym = sprintf "%04d%02d", $_[5] + 1900, $_[4] + 1;
	$logfile =~ s/%/$ym/;
    }

    @engines0 = sort('altavista', 'excite', 'fast', 'fresheye', 'goo',
		     'infoseek', 'lycos', 'nexearch', 'tocc', 'verno',
		     'yahoo');
    @engines1 = sort('altavista', 'excite', 'fast', 'fresheye', 'goo',
		     'infoseek', 'lycos', 'nexearch', 'tocc', 
		     'yahoo');
    $head_timeout = 10;

    if (! $ENV{'REQUEST_METHOD'}) {
	# Command line
	&getopts("he:s:m:SCt:bcD:");
	if ($opt_h || @ARGV == 0) {
	    print "Usage: $0 [options] word ...\n";
	    print "  -h         : Help\n";
	    print "  -e engines : Engines to use\n";
	    printf "\t\t(%s)\n", join(',', @engines0);
	    print "  -s {f,m,s} : Speed (fast, medium, slow)\n";
	    print "  -m n       : Max number of items (0 means all)\n";
	    print "  -S         : List by site\n";
	    print "  -C         : List with comments\n";
	    print "  -t target  : Target window name\n";
	    print "  -b         : Enable Bookmarks\n";
	    print "  -c         : Enable Check links\n";
	    print "  -D {s,l}   : Debug (save, load)\n";
	    exit 1;
	}
	$cgi = 0;
	$query = join(' ', @ARGV);
	if ($opt_e) {
	    foreach $e (split(/,/, $opt_e)) {
		next unless grep($e eq $_, @engines0);
		push(@engines, $e);
	    }
	}
	$speed = $opt_s;
	$max_show = $opt_m;
	$list_by_site = $opt_S;
	$with_comment = $opt_C;
	$target_win = $opt_t;
	$use_bm = $opt_b;
	$check_links = $opt_c;
	$debug = $opt_D;
    } else {
	# CGI
	$cgi = 1;
	&cgi_params;
	$query = $param{'q'};
	@engines = grep($param{$_}, @engines0);
	$speed = $param{'speed'};
	$max_show = $param{'max'};
	$list_by_site = $param{'site'};
	$with_comment = $param{'comment'};
	$target_win = $param{'tw'};
	$use_bm = $param{'bm'};
	$check_links = $param{'check'};
	$remote_host = $ENV{'REMOTE_HOST'} || '?';
	$remote_addr = $ENV{'REMOTE_ADDR'} || '?';
    }
    &jcode::convert(\$query, 'euc', undef, 'z');
    &jcode::tr(\$query, "\xa1\xa1", " ");
    $query = &remove_strange_chars($query);
    $query =~ s/^\s+//;
    $query =~ s/\s+$//;
    $q_euc = &url_encode($query);
    $q_sjis = &url_encode(&jcode::sjis($query, 'euc'));
    if (@engines == 0) {
	@engines = @engines1;
    }
    @engines = sort @engines;
    if ($speed eq 'm') {
	($max_hits, $send_timeout, $recv_timeout) = (30, 5, 20);
    } elsif ($speed eq 's') {
	($max_hits, $send_timeout, $recv_timeout) = (50, 10, 30);
    } else {
	$speed = 'f';
	($max_hits, $send_timeout, $recv_timeout) = (10, 3, 10);
    }
    $max_show = 20 unless defined $max_show;
    $max_show = int($max_show);
    $max_show = 20 if $max_show < 0;
    $max_check = 20;
    $list_by_site = int($list_by_site);
    $with_comment = int($with_comment);
    $target_win =~ tr/a-zA-Z0-9_//cd;
    if ($target_win) {
	$target = " target=$target_win";
    } else {
	$target = " target=_top";
    }
    if ($use_bm && $bmpage && $bmscript) {
	$use_bm = 1;
    } else {
	$use_bm = 0;
    }
    $check_links = int($check_links);
}

##
## Print Header
##
sub print_header {
    my $title0 = "Metcha Search for: ";
    my $q = &html_encode($query);
    my (%sel_speed, %sel_max, %sel_site, %sel_comment);
    $sel_speed{$speed} = ' selected';
    $sel_max{$max_show} = ' selected';
    $sel_site{$list_by_site} = ' selected';
    $sel_comment{$with_comment} = ' selected';
    $menu = "<a href=\"$toppage\">ȥåץڡ</a>";
    $menu .= " | <a href=\"$helppage\" target=_top>إ</a>";
    my ($bmsrc, $bmload, $bmwaring);
    if ($use_bm) {
	$bmsrc = &javascript($bmscript, "") . "\n";
	$bmbody = " onload=\"refreshCheck(document.results)\""
	    . " onfocus=\"refreshCheck(document.results)\"";
	$bmwarning = "<noscript><p><font color=#ff0000>"
	    . "JavaScriptCookie򥪥ˤʤȤ굡ǽѤǤޤ"
	    . "</font></p></noscript>\n";
	$menu .= " | <a href=\"$bmpage\" target=BOOKMARKS>ɽ</a>";
	my $js = "javascript:top.location.href=location.href";
	$menu .= " | <a href=\"$js\">ե졼</a>";
    }
    print "Content-type: text/html\n";
    if ($max_age) {
	print "Cache-Control: max-age=$max_age\n";
    }
    print "\n";
    $_ = <<EOM
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=EUC-JP">
<title>$title0$q</title>
$bmsrc</head>
<body bgcolor=#ffffff$bmbody>
<table border=0 cellpadding=0 cellspacing=0 bgcolor=#000060 width=100%>
<tr><td>&nbsp;&nbsp;<font color=#fffff><font face="sans-serif">$title0</font>$q</font></td></tr>
</table>
<br>
[ $menu ]
<br>
<form name="metcha" method="get" action="$metcha">
<input size=50 name="q" value="$q"><input type="submit" value="">
<br>
<select name="speed">
<option value="f"$sel_speed{'f'}>/®
<option value="m"$sel_speed{'m'}>/®
<option value="s"$sel_speed{'s'}>/®
</select>
<select name="max">
<option$sel_max{10}>10
<option$sel_max{20}>20
<option$sel_max{50}>50
<option$sel_max{100}>100
<option value="0"$sel_max{0}>٤
</select>
<select name="site">
<option value="0"$sel_site{0}>ڡ
<option value="1"$sel_site{1}>Ȥ
</select>
<select name="comment">
<option value="0"$sel_comment{0}>ʸʤ
<option value="1"$sel_comment{1}>ʸ
</select>
EOM
    ;
    print;
    foreach (@engines) {
	print "<input type=hidden name=$_ value=1>\n";
    }
    $_ = <<EOM
<input type=hidden name=tw value=\"$target_win\">
<input type=hidden name=bm value=$use_bm>
</form>
$bmwarning<hr>
EOM
    ;
    print;
}

##
## Print Links
##
sub print_links {
    my %links;
    my ($eng, $link);
    $links{'google'} = 'http://www.google.com/search?'
	. 'q=' . $q_sjis . '&hl=ja';
    $links{'kensaku'} = 'http://kensaku.org/search.cgi?version=4&lang=ja&'
	. 'key=' . $q_euc . '&number=100&sort=count';
    $links{'yahoo'} = 'http://search.yahoo.co.jp/bin/search?'
	. 'p=' . $q_euc;
    $links{'lycos'} = 'http://search.lycos.co.jp/main.html?'
	. 'query=' . $q_sjis . '&encoding=shift-jis';
    print "<hr>\n";
    print "<font size=-1>¾θ󥸥Ǹ: ";
    foreach $eng (sort keys %links) {
	print "[<a href=\"$links{$eng}\" target=_top>$eng</a>] ";
    }
    print "</font>\n";
}

##
## Print Trailer
##
sub print_trailer {
    print "<hr>\n";
    print "[ $menu ]\n";
    print "<hr>\n";
    print "<address>\nMaintained by $maintainer / ";
    print "Programmed by $programmer\n</address>\n";
    print "</body>\n</html>\n";
}

##
## GMT string
##
sub GMT_string {
    my ($time) = @_;
    @_ = split(/\s+/, scalor(gmtime($time)));
    sprintf "%s, %02d %s %s %s GMT", $_[0], $_[2], $_[1], $_[4], $_[3];
}

##
## List Good Pages
##
sub list_good_pages {
    my ($eng, $href, $title, @e, $e);
    print "<ul>\n";
    foreach $href (sort keys %good) {
	@e = sort keys %{$good{$href}};
	next if @e == 0;
	$title = $good{$href}{$e[0]};
	$title = $href unless $title;
	$e = join(', ', @e);
	print "<li> ڡ: ";
	&checkbox("g$good", $href, $title);
	print "<a href=\"$href\"$target>$title</a> ";
	print "<font size=-1 color=#808080>($e)</font>\n";
	++$good;
    }
    print "</ul>\n";
}

##
## List by page
##
sub list_by_page {
    my @hrefs = @_;
    my $href;
    print "<dl>\n";
    foreach $href (@hrefs) {
	&list_href($href);
    }
    print "</dl>\n";
}

##
## List by site
##
sub list_by_site {
    my @hrefs = @_;
    my ($site, @site_hrefs, $href);
    print "<dl>\n";
    foreach $site (sort by_site_max %site_max) {
	@site_hrefs = ();
	foreach $href (@hrefs) {
	    next if $href !~ m|^http://([\-\.\w]+)|i;
	    next if $1 ne $site;
	    push(@site_hrefs, $href);
	}
	next if @site_hrefs == 0;
	print "<dt> <i>$site</i>\n";
	print "<dd>\n";
	print "<dl>\n";
	foreach $href (@site_hrefs) {
	    &list_href($href);
	}
	print "</dl>\n";
    }
    print "</dl>\n";
}

##
## List URL
##
sub list_href {
    my ($href) = @_;
    my (@e, $eng, $e, $title, $comment, $total, $ml, $i, $date);
    foreach $eng (sort keys %{$score{$href}}) {
	push(@e, "$eng#$rank{$href}{$eng}=$score{$href}{$eng}");
    }
    $e = join(', ', @e);
    $i = $total_rank{$href};
    $title = ($title{$href} || $href);
    $comment = ($comment{$href} || $title);
    $total = int($total{$href});
    # $ml = "$metchalink?id=$id";
    # $ml .= "&url=" . &url_encode($href);
    print "<dt> $i ";
    print "<font size=-1 color=#ff0000>[$total]</font> ";
    &checkbox("p$i", $href, $title);
    print "<a href=\"$href\"$target>$title</a>\n";
    if ($check_links) {
	&check_link($href);
	$date = $date{$href};
	print "<font size=-1>(";
	if ($date eq 'err') {
	    print "Not Found";
	} else {
	    if ($date) {
		@_ = split(/[-,:\s]+/, $date);
		$_[3] += 1900 if $_[3] < 1900;
		$date = "$_[1] $_[2] $_[3]";
	    }
	    printf "Last-Modified: %s, Size %s bytes",
	    	($date || '?'), ($len{$href} || '?');
	}
	print ")</font>\n";
    }
    if ($with_comment) {
	print "<dd> <font size=-1>$comment</font>\n";
    }
    print "<dd> <font size=-1 color=#808080>($e) <i>$href</i></font>\n";
    # print "[<a href=\"$ml\">is referred from</a>]\n";
}

##
## Checkbox
##
sub checkbox {
    my ($name, $href, $title) = @_;
    if ($use_bm) {
#	$_ = "$href " . &url_encode($title);
#	print "<input type=checkbox name=\"$name\" value=\"$href $title\"> ";
	print "<input type=checkbox ",
		"onclick=\"toBM(this)\" value=\"$href $title\"> ";
#		"onclick=\"toBM(this)\" value=\"$_\"> ";
    }
}

##
## JavaScript block
##
sub javascript {
    my ($src, $prog) = @_;
    if ($src) {
	$src = " src=\"$src\"";
    }
    "<script language=\"javascript\"$src>$prog</script>";
}

##
## Extract URLs from $content{$eng}
## and set the followings
##	$length{$eng} : length of $content{$eng}
##	$count{$eng}  : number of URLs
##	$sum{$eng}    : sum of URL scores
##	$max{$eng}    : max of URL scores
##	$min{$eng}    : min of URL scores
##	$score{$href}{$eng} : score of URL
##	$title{$href}       : title of URL
##
sub extract {
    my ($eng) = @_;
    $_ = $content{$eng};
    $length{$eng} = length($_);
    tr/\r\n/  /;
    &jcode::convert(\$_, 'euc', undef, 'z');
    $count{$eng} = 0;
    $sum{$eng} = 0;
    # call the extract program of $eng
    &{$extract{$eng}}($eng, join(' ', $_));
    $max{$eng} = int($max{$eng});
    $min{$eng} = int($min{$eng});
}

sub add_url {
    my ($eng, $href, $title, $score, $comment) = @_;
    my $code;
    if ($href !~ m|^http://|) {
	return;
    }
    $href = &remove_strange_chars($href);
    $href =~ s/\s+$//;
    $href =~ s/ /%20/g;
    $href =~ s/"/%22/g;
    $href =~ s/'/%27/g;
    $href =~ s/</%3c/g;
    $href =~ s/>/%3e/g;
    $href =~ s/%7e|%7E/~/g;
    $href =~ s|^(http://[\w_!*\$\-\.\+\'\(\)]+)\.([:/])|$1$2|i;
    $href =~ s|^(http://[\w_!*\$\-\.\+\'\(\)]+):80\b|$1|i;
    if ($href =~ m|^(.*)/$|) {
	if (defined $score{$1} && ! defined $score{$href}) {
	    $score{$href} = $score{$1}; delete $score{$1};
	    $title{$href} = $title{$1}; delete $title{$1};
	    $comment{$href} = $comment{$1}; delete $comment{$1};
	    $rank{$href} = $rank{$1}; delete $rank{$1};
	    if ($good{$1}) {
		$good{$href} = $good{$1}; delete $good{$1};
	    }
	}
    } else {
	if (defined $score{"$href/"}) {
	    $href = "$href/";
	}
    }
    if (defined $score{$href}{$eng}) {
	return;
    }
    if ($score eq 'good') {
	$good{$href}{$eng} = $title;
	return;
    }
    if (! defined $title{$href} && $title) {
	$code = &jcode::getcode(\$title);
	if (! $code || $code eq 'euc') {
	    $title = &html_text($title);
	    if (length($title{$href}) < length($title)) {
	        $title{$href} = $title;
            }
	}
    }
    if (! defined $comment{$href} && $comment) {
	$code = &jcode::getcode(\$comment);
	if (! $code || $code eq 'euc') {
	    $comment = &html_text($comment);
	    if (length($comment{$href}) < length($comment)) {
		$comment{$href} = $comment;
	    }
	}
    }
    ++$count{$eng};
    $rank{$href}{$eng} = $count{$eng};
    $score{$href}{$eng} = $score;
    $sum{$eng} += $score;
    if (! $max{$eng} || $score > $max{$eng}) {
	$max{$eng} = $score;
    }
    if (! $min{$eng} || $score < $min{$eng}) {
	$min{$eng} = $score;
    }
}

##
## Calculate the total score
##
sub calc_total {
    my ($href, $eng, $total, $site);
    my $t = 'TOTAL';
    $count{$t} = 0;
    $sum{$t} = 0;
    foreach $href (keys %score) {
	next if %{$score{$href}} == 0;
	$total = 0;
	foreach $eng (@engines) {
	    next if $max{$eng} == 0;
	    next if $score{$href}{$eng} == 0;
	    $total += ($score{$href}{$eng} / $max{$eng} + 0.3) / 1.3;
	}
	$total = 100 * $total / $engines;
	$total{$href} = $total;
	++$count{$t};
	$sum{$t} += $total;
	if (! $max{$t} || $total > $max{$t}) {
	    $max{$t} = $total;
	}
	if (! $min{$t} || $total < $min{$t}) {
	    $min{$t} = $total;
	}
	if ($href !~ m|^http://([\w_!*\$\-\.\+\'\(\)]+)|i) {
	    next;
	}
	$site = $1;
	if (! $site_max{$site} || $total > $site_max{$site}) {
	    $site_max{$site} = $total;
	}
    }
    $max{$t} = int($max{$t});
    $min{$t} = int($min{$t});
}

sub by_total {
    ($total{$b} <=> $total{$a}) || ($a cmp $b);
}

sub by_site_max {
    ($site_max{$b} <=> $site_max{$a}) || ($a cmp $b);
}

##
## Request to a search engine
##
sub request {
    local ($eng) = @_;
    local $sock = "S$eng";
    local $server;
    $SIG{'ALRM'} = \&request_timeout;
    alarm($send_timeout);
eval <<'EOM';
    $server = ($ip{$eng} || $server{$eng});
    if (&tcp_client($sock, $server, $port{$eng})) {
	print $sock "GET $request{$eng} HTTP/1.0\015\012",
		    "User-Agent: $agent_name\015\012",
#	            "Accept-Language: ja,en\015\012",
#	            "Accept-Charset: euc-jp\015\012",
	            "Referer: $toppage\015\012",
	            "From: $from\015\012\015\012";
	$sock{$eng} = $sock;
	alarm(0);
    }
EOM
    alarm(0);
    $SIG{'ALRM'} = IGNORE;
    if (! $sock{$eng}) {
	$status{$eng} = 'err';
	close($sock);
    }
}

sub request_timeout {
    die;
}

##
## Receive results from all search engines
##
sub receive_all {
    my ($time0) = @_;
    my ($eng, $sock);
    my ($c, $r_bitset, $nfound, $timeleft);
    my ($buf, $bytes);
    my $buflen = 8192;
    my $wait = 2;
    foreach $eng (@engines) {
	$sock = $sock{$eng};
	next unless $sock;
	# Non blocking IO
	fcntl($sock, F_SETFL, O_NONBLOCK);
    }
    while (1) {
	$c = 0;
	$r_bitset = '';
	foreach $eng (@engines) {
	    $sock = $sock{$eng};
	    next unless $sock;
	    next if $status{$eng} eq 'ok';
	    $status = $status{$eng};
	    vec($r_bitset, fileno($sock), 1) = 1;
	    ++$c;
	}
	last if $c == 0;
	($nfound, $timeleft) = select($r_bitset, undef, undef, $wait);
	last if $nfound < 0;	# Error
	if ($nfound == 0) {	# Timeout
	    last if time - $time0 > $recv_timeout;
	    next;
	}
	foreach $eng (@engines) {
	    $sock = $sock{$eng};
	    next unless $sock;
	    next if $status{$eng} eq 'ok';
	    next if vec($r_bitset, fileno($sock), 1) == 0;
	    $bytes = sysread($sock, $buf, $buflen);
	    if ($bytes == 0) {
		# closed
		$status{$eng} = 'ok';
		close($sock);
		$time{$eng} = time - $time0;
		print "$eng=$time{$eng}\n";
	    } else {
		$content{$eng} .= $buf;
	    }
	}
    }
    foreach $eng (@engines) {
	$sock = $sock{$eng};
	next unless $sock;
	next if $status{$eng} eq 'ok';
	close($sock);
	$status{$eng} = 'x';	# Timeout
	$time{$eng} = time - $time0;
	print "$eng=x\n";
    }
}

##
## Save content to a file for debug
##
sub save_content {
    my ($eng, $file) = @_;
    open (FILE, ">$file") || die "$0: $file $!";
    $_ = $content{$eng};
    # tr/\r/ /d;
    # &jcode::convert(\$_, 'euc', undef, 'z');
    print FILE $_;
    close(FILE);
}

##
## Load content from a file for debug
##
sub load_content {
    my ($eng, $file) = @_;
    if (! open (FILE, "<$file")) {
	return;
    }
    @_ = <FILE>;
    $content{$eng} = join('', @_);
    close(FILE);
}

##
## Check link
##
sub check_link {
    local ($href) = @_;
    local ($host, $port, $query, $ok);
    local $sock = "S";
    if ($href !~ m|^http://([\-\.\w]+)|i) {
	return;
    }
    $host = $1;
    $port = 80;
    $query = $';
    if ($query =~ /^:(\d+)/) {
	$port = $1;
	$query = $';
    }
    $SIG{'ALRM'} = \&check_link_timeout;
    alarm($head_timeout);
eval <<'EOM';
    if (&tcp_client($sock, $host, $port)) {
	print $sock "HEAD $query HTTP/1.0\015\012",
	            "Host: $host:$port\015\012",
	            "User-Agent: $agent_name\015\012",
		    "Referer: $toppage\015\012",
	            "From: $from\015\012\015\012";
	$_ = <$sock>;
	if (! /^HTTP\S*\s+(2|3)/i) {
	    $date{$href} = 'err';
	} else {
	    while (<$sock>) {
		tr/\r\n//d;
		last if $_ eq '';
		if (/^Location:\s*(\S.*)\s*$/i) {
		    $location{$href} = $1;
		} elsif (/^Last-modified:\s*(\S.*)\s*$/i) {
		    $date{$href} = $1;
		} elsif (/^Content-length:\s*(\d+)\s*$/i) {
		    $len{$href} = $1;
		}
	    }
	}
    }
EOM
    alarm(0);
    $SIG{'ALRM'} = IGNORE;
    close($sock);
}

sub check_link_timeout {
    die;
}

##
## TCP client
##
sub tcp_client {
    my ($sock, $host, $port) = @_;
    my ($addr, $proto, $ent, $oldh);
    @_ = gethostbyname($host);
    $addr = $_[4];
    $proto = getprotobyname('tcp');
    if ($port !~ /^\d+$/) {
	@_ = getservbyname($port, 'tcp');
	$port = $_[2];
    }
    socket($sock, PF_INET, SOCK_STREAM, $proto);
    $ent = pack('S n a4 x8', AF_INET, $port, $addr);
    if (! connect($sock, $ent)) {
	return undef;
    }
    $oldh = select;
    select($sock); $| = 1; select($oldh);
    $sock;
}

##
## HTML text
##
sub html_text {
    $_ = shift;
    $_ = &remove_strange_chars($_);
    s/^\s+//;
    s/\s+$//;
    s|<.*?>||g;
    s/</&lt;/g;
    s/>/&gt;/g;
    s/"/&quot;/g;
    $_;
}

##
## HTML encode
##
sub html_encode {
    $_ = shift;
    s/&/&amp;/g;
    s/</&lt;/g;
    s/>/&gt;/g;
    s/"/&quot;/g;
    $_;
}

##
## URL encode
##
sub url_encode {
    $_ = shift;
    s/([^ \w])/'%'.uc(unpack('H2',$1))/ge;
    s/\s/+/g;
    $_;
}

##
## Remove strange characters (non-ASCII, non-EUC)
##
sub remove_strange_chars {
    $_ = shift;
    my $r;
    while (/([\x20-\x7e]|\x8f?[\xa1-\xfe][\xa1-\xfe])+/g) {
	$r .= $&;
    }
    $r;
}

##
## Get CGI parameters
##
sub cgi_params {
    my ($args, $tag, $value);
    if ($ENV{'REQUEST_METHOD'} eq 'POST'){
	sysread(STDIN, $args, $ENV{'CONTENT_LENGTH'});
    } else {
	$args = $ENV{'QUERY_STRING'};
    }
    foreach (split('&', $args)) {
	($tag, $value) = split(/=/, $_, 2);
	$value =~ s/\+/ /g;
	$value =~ s/%(..)/pack('c', hex($1))/ge;
	$param{$tag} = $value;
    }
}

##
## Write log
##
sub write_log_1 {
    my ($log, $id, $t) = @_;
    my ($line, $eng);
    foreach $eng (@engines) {
	$line .= "$eng ";
    }
    $line .= "speed=$speed max=$max_show site=$list_by_site ";
    $line .= "comment=$with_comment ";
    $line .= "tw=$target_win bm=$use_bm check=$check_links ";
    $line .= "q=$query";
    open(LOG, ">>$log") || die "cannot append $log\n";
    flock(LOG, 2);		# exclusive lock
    seek(LOG, 0, 2);
    print LOG "Q $id $t $remote_host $remote_addr $line\n";
    # flock(LOG, 8);		# unlock
    close(LOG);
    if ($logmod) { chmod $logmod, $log };
}

sub write_log_2 {
    my ($log, $id, $t) = @_;
    my ($line, $eng);
    foreach $eng (@engines) {
	if ($status{$eng} eq 'err') {
	    $line .= "$eng=$status{$eng} ";
	} else {
	    $line .= "$eng=$status{$eng},$time{$eng},$count{$eng},";
	    $line .= "$max{$eng}-$min{$eng} ";
	}
    }
    $eng = 'TOTAL';
    $line .= "$eng=$status{$eng},$time{$eng},$count{$eng},";
    $line .= "$max{$eng}-$min{$eng} ";
    $line .= "good=$good ";
    $line .= "speed=$speed max=$max_show site=$list_by_site ";
    $line .= "comment=$with_comment ";
    $line .= "tw=$target_win bm=$use_bm check=$check_links ";
    $line .= "q=$query";
    open(LOG, ">>$log") || die "cannot append $log\n";
    flock(LOG, 2);		# exclusive lock
    seek(LOG, 0, 2);
    print LOG "R $id $t $remote_host $remote_addr $line\n";
    # flock(LOG, 8);		# unlock
    close(LOG);
    # if ($logmod) { chmod $logmod, $log };
}

##
## Search engines
##
sub init_engine {
    my ($eng, $max_hits) = @_;
    my ($server, $ip, $port, $r, $extract);
    $port = 80;
    $max_hits = 100 if $max_hits > 100;
    if ($eng eq 'altavista') {
	$server = 'www.altavista.com';
	$ip = '209.73.164.95';
        $r = '/sites/search/web?q=' . $q_sjis;
	$r .= '&pg=q&kl=ja&enc=sjis&Translate=on';
	$r .= '&nbq=' . $max_hits;
	$extract = \&altavista;
    } elsif ($eng eq 'excite') {
	$server = 'www.excite.co.jp';
	$ip = '198.3.103.129';
	$max_hits = 50 if $max_hits > 50;
	$r = '/search.gw?look=excite_jp&target=web&showSummary=false&';
	$r .= 'search=' . $q_sjis;
	$r .= '&perPage=' . $max_hits;
	$extract = \&excite;
    } elsif ($eng eq 'fast') {
	$server = 'www.alltheweb.com';
	$ip = '66.77.74.20';
	$r = '/search?cat=web&lang=japanese&';
	$r .= 'query=' . $q_sjis;
	$r .= '&hits=' . $max_hits;
	$extract = \&fast;
    } elsif ($eng eq 'fresheye') {
	$server = 'search.fresheye.com';
	$ip = '203.139.121.226';
	$r = '/?kw=' . $q_euc;
	$extract = \&fresheye;
    } elsif ($eng eq 'goo') {
	$server = 'www.goo.ne.jp';
	$ip = '210.150.25.37';
	$max_hits = 25 if $max_hits == 30;
	$r = '/default.asp?MT=' . $q_euc;
	$r .= '&SM=MC&DC=' . $max_hits;
	$r .= '&SDB=PG';
	$extract = \&goo;
    } elsif ($eng eq 'infoseek') {
	$server = 'www.infoseek.co.jp';
	$ip = '206.3.30.196';
	$r = '/Titles?col=JW&qt=' . $q_euc;
	$r .= '&sv=JP&lk=&nh=' . $max_hits;
	$r .= '&tds=0&qp=0&search=%2F&rf=0&ud4=1';
	$extract = \&infoseek;
    } elsif ($eng eq 'lycos') {
	$server = 'wisenut.lycos.co.jp';
	$ip = '202.214.130.41';
	$r = '/?q=' . $q_sjis;
	$r .= '&ie=shift-jis';
	$r .= '&c=' . $max_hits;
	$extract = \&lycos;
    } elsif ($eng eq 'nexearch') {
	$server = 'search.naver.co.jp';
	$ip = '202.234.172.32';
	$r = '/search.naver?query=' . $q_euc;
	$extract = \&nexearch;
    } elsif ($eng eq 'tocc') {
	$server = 'www.tocc.co.jp';
	$ip = '210.172.228.51';
	$r = '/search/servlet/SearchServlet?QRY=' . $q_euc;
	$r .= '&CCM=M%3AW&NRS=' . $max_hits;
	$r .= '&UNI=HST&ADP=0.0&FIL=ABS%2CDAT%2CSCR%2CTTL%2CURL';
	$r .= '&FRS=0&UD0=SMP&UNI=DOC';
	$extract = \&tocc;
    } elsif ($eng eq 'verno') {
	$server = 'verno.ueda.info.waseda.ac.jp';
	$ip = '133.9.215.87';
	$r = '/verno-099.cgi?index-and=' . $q_euc;
	$r .= '&kosuu=' . $max_hits;
	$r .= '&dig=-1';
	$extract = \&verno;
    } elsif ($eng eq 'yahoo') {
	$server = 'google.yahoo.co.jp';
	$ip = '206.3.4.197';
	$r = '/bin/query?p=' . $q_euc;
	$r .= '&n=' . $max_hits;
	$extract = \&yahoo;
    }
    if ($proxy_server && $proxy_port && $server !~ /$no_proxy/) {
	$r = "http://$server" . $r;
	$server = $proxy_server;
	undef $ip;
	$port = $proxy_port;
    }
    $server{$eng} = $server;
    $ip{$eng} = $ip;
    $port{$eng} = $port;
    $request{$eng} = $r;
    $extract{$eng} = $extract;
}

sub altavista {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|^.*?We found.*?results:|i) {
	return;
    }
    $content = $';
    $content =~ s|<p class=\"s\">.*$||i;
    $re = '<a\s+href="/r\?ck_sm=.*?status=\'(http.*?)\'.*?>(.*?)</a>\s*';
    $re .= '</span><br>\s*(.*?)<span';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$title =~ s|</?b>||g;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub excite {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|^.*?<ol.*?>|i) {
	return;
    }
    $content = $';
    $content =~ s|</ol.*$||i;
    $re = '<li><a href="(.*?)">(.*?)</a>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$title =~ s|</?b>||g;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub fast {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|^.*?<dl class=mainresult>|i) {
	return;
    }
    $content = $';
    $content =~ s|</dl>.*$||i;
    $re = '<a .*?href="/go/\d+/H/web/http/(.*?)".*?>(.*?)</a>';
    $re .= '</dt>\s*<dd>(.*?)<br>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = "http://$1";
	$title = $2;
	$comment = $3;
	if ($comment =~ /^<a href/) {
	    $comment = undef;
	}
	$title =~ s|</?b>||g;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub fresheye {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    if ($content =~ m|ѡ<BR>\s*Х|i) {
	$href = "http://$server{$eng}$request{$eng}";
	$title = "ѡХ";
	&add_url($eng, $href, $title, 'good');
    }
    if ($content =~ m|<A HREF=(\S+from=zuba)>(.*?)</A>|i) {
	$href = $1;
	$title = $2;
	if ($href =~ m|^http://search.fresheye.com/\?go=(.*?)&from=zuba|) {
	    $href = $1;
	    $href =~ s/%(..)/pack('c', hex($1))/ge;
	} else {
	    $href = "http://www.fresheye.com$href";
	}
	&add_url($eng, $href, $title, 'good', undef);
    }
    #$content =~ s|^.*?<!-- start url -->||i;
    #$content =~ s|<!-- end url -->.*$||i;
    $re = '<B>(\d+)%</B>\s*</TD>\s*<TD>\s*</TD>\s*<TD>\s*';
    $re .= '<A HREF="?http://search.fresheye.com/\?go=(.*?)"?>';
    $re .= '(.*?)</A><BR><font.*?>.*?</font><br>(.*?)<br>';
    while ($content =~ /$re/igo) {
	$href = $2;
	$title = $3;
	$comment = $4;
	$score = $1;
	$href =~ s/%3f/?/ig;
	&add_url($eng, $href, $title, $score, $comment);
    }
}

sub goo {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|^.*?<!-- result -->|i) {
	return;
    }
    $content = $';
    $content =~ s|<!-- result/.*$||i;
    $re = '<b>\d+</b>\s*<a href.*?><img.*?></a>';
    $re .= '<a href.*?>(.*?)</a><br>\s*';
    $re .= '<div.*?>\s*<font.*?>(.*?)</font><br>\s*';
    $re .= '<font.*?>(.*?)</font>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $2;
	$title = $1;
	$comment = $3;
	$score = int(100 - 30*$n/$max_hits);
	$title =~ s|</?b>||gi;
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub infoseek {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|^.*?<!-- START ResURL_Search -->|i) {
	return;
    }
    $content = $';
    $content =~ s|<!-- END ResURL_Search -->.*$||i;
    $re = '<a href="(.*?)">(.*?)</a>.*?';
    $re .= '<font.*?>\s*(\d+)%';
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$score = $3;
	&add_url($eng, $href, $title, $score, $comment);
    }
}

sub lycos {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|<ol>|i) {
	return;
    }
    $content = $';
    $content =~ s|</ol>.*$||i;
    $re = '<li.*?><!--(\d+)--><a href=\"(.*?)\">(.*?)</a>';
    $re .= '<br><small>(.*?)<br>';
    while ($content =~ /$re/igo) {
	$href = $2;
	$title = $3;
	$comment = $4;
	$score = int(100 - 30*$1/$max_hits);
	$title =~ s|</?b>||gi;
	&add_url($eng, $href, $title, $score, $comment);
    }
}

sub nexearch {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    $re = 'icn_shortcut.gif.*?<a href=(\S+).*?>\'<b>(.*?)</b>';
    if ($content =~ /$re/i) {
	$href = $1;
	$title = $2;
	&add_url($eng, $href, $title, 'good');
    }
    unless ($content =~ m|^.*?<ul>|i) {
	return;
    }
    $content = $';
    $content =~ s|<center>.*$||i;
    $re = '<td><small>\s*<a href="(.*?)".*?>(.*?)</a></small></td>';
    $re .= '.*?<font size=2.*?>(.*?)<(br|font)';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub tocc {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ s|^.*?<!-- Web Search Result Items -->||i) {
	return;
    }
    $content = $';
    $content =~ s|</ol>.*$||i;
    $re = '<li>\s*<a href="(.*?)">(.*?)</a>\s*';
    $re .= '<br\s*/>\s*<font.*?>(.*?)</font>\s*<br\s*/>\s*<font.*?>\s*(\d+)%';
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$score = $4;
	$title =~ s|</?b>||gi;
	&add_url($eng, $href, $title, $score, $comment);
    }
}

sub verno {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ s|^.*?</FORM>||i) {
	return;
    }
    $re = ' (\d+)\s*pt\.\s*';
    $re .= '<a href="(.*?)">(.*?)</a>';
    while ($content =~ /$re/igo) {
	$href = $2;
	$title = $3;
	$score = $1;
	&add_url($eng, $href, $title, $score, $comment);
    }
}

sub yahoo {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    unless ($content =~ m|^.*?<ul>|i) {
	return;
    }
    $content = $';
    $content =~ s|</ul>.*$||i;
    $re = '<li><a href=\"(.*?)\">(.*?)</a>';
    $re .= '<br><small>(.*?)<br>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$title =~ s|</?b>||g;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

